from pyspedas.utilities.dailynames import dailynames
from pyspedas.utilities.download import download
from pyspedas.tplot_tools import time_clip as tclip
from pyspedas.tplot_tools import cdf_to_tplot

from .config import CONFIG


def load(trange=['2013-11-5', '2013-11-6'], 
         instrument='cindi',
         datatype='efield_1sec',
         prefix='',
         suffix='', 
         get_support_data=False, 
         varformat=None,
         varnames=[],
         downloadonly=False,
         notplot=False,
         no_update=False,
         time_clip=False,
         force_download=False):
    """
    This function loads data from the CNOFS mission; this function is not meant
    to be called directly; instead, see the wrappers:
        pyspedas.projects.cnofs.cindi
        pyspedas.projects.cnofs.plp
        pyspedas.projects.cnofs.vefi
    """

    if prefix is None:
        prefix=''
    if suffix is None:
        suffix=''

    if instrument == 'cindi':
        pathformat = instrument+'/ivm_500ms_cdf/%Y/cnofs_'+instrument+'_ivm_500ms_%Y%m%d_v??.cdf'
    elif instrument == 'plp':
        pathformat = instrument+'/plasma_1sec/%Y/cnofs_'+instrument+'_plasma_1sec_%Y%m%d_v??.cdf'
    elif instrument == 'vefi':
        pathformat = instrument+'/'+datatype+'/%Y/cnofs_'+instrument+'_'+datatype+'_%Y%m%d_v??.cdf'

    # find the full remote path names using the trange
    remote_names = dailynames(file_format=pathformat, trange=trange)

    out_files = []

    files = download(remote_file=remote_names, remote_path=CONFIG['remote_data_dir'], local_path=CONFIG['local_data_dir'], no_download=no_update, force_download=force_download)
    if files is not None:
        for file in files:
            out_files.append(file)

    out_files = sorted(out_files)

    if downloadonly:
        return out_files

    tvars = cdf_to_tplot(out_files, prefix=prefix, suffix=suffix, get_support_data=get_support_data, varformat=varformat, varnames=varnames, notplot=notplot)
    
    if notplot:
        return tvars

    if time_clip:
        for new_var in tvars:
            tclip(new_var, trange[0], trange[1], suffix='')

    return tvars



"""Test plotting functions (mostly for pseudovariables)"""

import os
import unittest
import numpy as np
from pyspedas.projects import themis
from pyspedas import (
    store_data,
    get_data,
    set_coords,
    set_units,
    cotrans,
    tplotxy,
    tplotxy3,
    tplotxy3_add_mpause,
    tplotxy3_add_neutral_sheet,
    tkm2re,
    tplot_restore,
    tplot_names,
)
from pyspedas.utilities.config_testing import TESTING_CONFIG
from pyspedas import bshock_2, mpause_2, neutral_sheet

# Whether to display plots during testing
global_display = TESTING_CONFIG["global_display"]
# Directory to save testing output files
output_dir = TESTING_CONFIG["local_testing_dir"]
# Ensure output directory exists
save_dir = os.path.join(output_dir, "utilities")
if not os.path.exists(save_dir):
    os.makedirs(save_dir)

default_trange = ["2020-01-01", "2020-01-02"]
global_display=False

class PlotTestCases(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        themis.state(trange=default_trange, probe=['a','b','c','d','e'])
        cls.bs = bshock_2()
        cls.mp = mpause_2()
        # This is time-dependent, so we'll just pick the midpoint of the first tplot variable
        d=get_data('tha_pos_gse')
        mid_time = (d.times[-1] - d.times[0])/2.0
        ns_x_re = -1.0*np.arange(0.0,375.0, 5.0)
        times=np.zeros(len(ns_x_re))
        times[:] = mid_time
        ns_gsm_pos=np.zeros((len(ns_x_re),3))
        ns_gsm_pos[:,0] = ns_x_re
        ns = neutral_sheet(times, ns_gsm_pos, model="aen", sc2NS=False)
        ns_gsm_pos[:,2] = ns
        store_data('ns_gsm_pos', data={'x':times, 'y':ns_gsm_pos})
        set_coords('ns_gsm_pos','GSM')
        set_units('ns_gsm_pos', 're')
        cotrans('ns_gsm_pos', 'ns_gse_pos',coord_in='gsm', coord_out='gse')
        cls.gse_dat = get_data('ns_gse_pos')

        tplot_restore('/tmp/ttrace_iono_t89.tplot')
        tplot_names()



    def test_themis_orbit_re(self):
        # Test a simple THEMIS orbit plot in the XY plane, plot units in re
        tplotxy('th?_pos_gse', title='THEMIS+ARTEMIS orbit plot 2020-01-01, units re', reverse_x=True, reverse_y=True, plot_units='re', display=global_display, save_png='orbits_thm_art_re.png')


    def test_themis_orbit_re_legends(self):
        # Test a simple THEMIS orbit plot in the XY plane with legends, plot units in re
        tplotxy('th?_pos_gse',title='THEMIS+ARTEMIS orbit plot 2020-01-01 with legends', legend_names=['THEMIS-A','THEMIS-B','THEMIS-C', 'THEMIS-D', 'THEMIS-E'], reverse_x=True, reverse_y=True, plot_units='re', display=global_display, save_png='orbits_thm_art_re_legends.png')

    def test_themis_orbit_re_badlegends(self):
        # Test a simple THEMIS orbit plot in the XY plane with bad legends, plot units in re
        tplotxy('th?_pos_gse', title='THEMIS+ARTEMIS orbit plot 2020-01-01 with bad legends',legend_names=['THEMIS-A','THEMIS-B'], reverse_x=True, reverse_y=True, plot_units='re', display=global_display, save_png='orbits_thm_art_re_badlegends.png')

    def test_themis_orbit_re_xz(self):
        # Test a simple THEMIS orbit plot in the XZ plane, plot units in re
         tplotxy('th?_pos_gse', title='THEMIS+ARTEMIS orbit plot 2020-01-01, XZ plane, units re',reverse_x=True, reverse_y=False, plane='xz', plot_units='re', display=global_display, save_png='orbits_thm_art_xz_re.png')

    def test_themis_orbit_re_yz(self):
        # Test a THEMIS orbit plot in the YZ plane, no X reversal, plot units in re
         tplotxy('th?_pos_gse', title='THEMIS+ARTEMIS orbit plot 2020-01-01, YZ plane, no reversal, units km', reverse_x=False, reverse_y=False, plane='yz', plot_units='re', display=global_display, save_png='orbits_thm_art_yz_noreverse_re.png')

    def test_themis_orbit_km(self):
        # Test a simple THEMIS-ARTEMIS orbit plot in the XY plane,plot units in km
        tplotxy('th?_pos_gse', title='THEMIS+ARTEMIS orbit plot 2020-01-01, units km',reverse_x=True, reverse_y=True, plot_units='km', display=global_display, save_png='orbits_thm_art_km.png')

    def test_themis_orbit_inner_mixedunits_km(self):
        # Test a simple THEMIS orbit plot in the XY plane, inner probes, plot units km, mixed input units
        tkm2re('the_pos_gse', km=False)
        tplotxy(['tha_pos_gse', 'thd_pos_gse', 'the_pos_gse_re'],title='THEMIS inner orbit plot 2020-01-01, YZ plane, no reversal, units km', reverse_x=True, reverse_y=True, plot_units='km', display=global_display, save_png='orbits_thm_mixed_km.png')

    def test_themis_orbit_inner_mixedunits_re(self):
        # Test a simple THEMIS orbit plot in the XY plane, inner probes, plot units re, mixed input units
        tkm2re('the_pos_gse', km=False)
        tplotxy(['tha_pos_gse', 'thd_pos_gse', 'the_pos_gse_re'],title='THEMIS inner orbit plot 2020-01-01, YZ plane, no reversal, units re', reverse_x=True, reverse_y=True, plot_units='re', display=global_display, save_png='orbits_thm_mixed_re.png')

    def test_themis_orbit_inner_linestyles(self):
        # Test a simple THEMIS orbit plot in the XY plane, inner probes, linestyles, plot units re, mixed input units
        tplotxy(['tha_pos_gse', 'thd_pos_gse', 'the_pos_gse'],title='THEMIS inner orbit plot 2020-01-01, YZ plane, linestyle, units re', legend_names=['THEMIS-A','THEMIS-B','THEMIS-C', 'THEMIS-D', 'THEMIS-E'], reverse_x=True, reverse_y=True, plot_units='re', linestyles=['solid','dotted','dashdot'], display=global_display, save_png='orbits_thm_linestyles.png')

    def test_themis_orbit_inner_markers(self):
        # Test a simple THEMIS orbit plot in the XY plane, inner probes, plot units re, mixed input units
        tplotxy(['tha_pos_gse', 'thd_pos_gse', 'the_pos_gse'],title='THEMIS inner orbit plot 2020-01-01, YZ plane, markers, units re', legend_names=['THEMIS-A','THEMIS-D', 'THEMIS-E'], reverse_x=True, reverse_y=True, plot_units='re', markers=['x','.','+'], markevery=60, display=global_display, save_png='orbits_thm_markers.png')

    def test_themis_orbit_inner_startendmarkers(self):
        # Test a simple THEMIS orbit plot in the XY plane, inner probes, plot units re, mixed input units
        tplotxy(['tha_pos_gse', 'thd_pos_gse', 'the_pos_gse'],title='THEMIS inner orbit plot 2020-01-01, YZ plane, no reversal, startend markers units re', reverse_x=True, reverse_y=True, plot_units='re', startmarkers=['x','.','+'], endmarkers=['x','.','+'], markers=None, display=global_display, save_png='orbits_thm_startendmarkers.png')

    def test_themis_orbit_inner_startmarkers(self):
        # Test a simple THEMIS orbit plot in the XY plane, inner probes, plot units re, mixed input units
        tplotxy(['tha_pos_gse', 'thd_pos_gse', 'the_pos_gse'], title='THEMIS inner orbit plot 2020-01-01, YZ plane, start markers, units re',reverse_x=True, reverse_y=True, plot_units='re', startmarkers=['x','.','+'], markers=None, display=global_display, save_png='orbits_thm_startmarkers.png')

    def test_themis_orbit_inner_linewidths(self):
        # Test a simple THEMIS orbit plot in the XY plane, inner probes, plot units re, mixed input units
        tplotxy(['tha_pos_gse', 'thd_pos_gse', 'the_pos_gse'],title='THEMIS inner orbit plot 2020-01-01, YZ plane, linewidths, units re', reverse_x=True, reverse_y=True, plot_units='re', linewidths=[1,2,3], display=global_display, save_png='orbits_thm_linewidths.png')

    def test_themis_orbit_inner_showearth_xy(self):
        # Test a simple THEMIS orbit plot in the XY plane, inner probes, plot units re, mixed input units
        tplotxy(['tha_pos_gse', 'thd_pos_gse', 'the_pos_gse'], title='THEMIS inner orbit plot 2020-01-01, XY plane, input km, units re',reverse_x=True, reverse_y=True, plot_units='re', show_centerbody=True, display=global_display, save_png='orbits_thm_showearth_xy.png')

    def test_themis_orbit_inner_showearth_xz(self):
        # Test a simple THEMIS orbit plot in the XY plane, inner probes, plot units re, mixed input units
         tplotxy(['tha_pos_gse', 'thd_pos_gse', 'the_pos_gse'],title='THEMIS inner orbit plot 2020-01-01, XZ plane, input km, units re', plane='xz', reverse_x=True, reverse_y=False, plot_units='re', show_centerbody=True, display=global_display, save_png='orbits_thm_showearth_xz.png')


    def test_themis_orbit_inner_showearth_yz(self):
        # Test a simple THEMIS orbit plot in the XY plane, inner probes, plot units re, mixed input units
         tplotxy(['tha_pos_gse', 'thd_pos_gse', 'the_pos_gse'],title='THEMIS inner orbit plot, YZ plane, no reversal, units re', plane='yz', plot_units='re', show_centerbody=True, display=global_display, save_png='orbits_thm_showearth_yz.png')

    def test_themis_orbit_inner_showearth_yz_nocenter(self):
        # Test a simple THEMIS orbit plot in the XY plane, inner probes, plot units re, mixed input units
        tplotxy(['tha_pos_gse', 'thd_pos_gse', 'the_pos_gse'],title='THEMIS inner orbit plot 2020-01-01, YZ plane, no reversal, no centering, units re', plane='yz', center_origin = False, plot_units='re', show_centerbody=True, display=global_display, save_png='orbits_thm_showearth_yz_nocenter.png')

    def test_themis_orbit_re_legends_3planes(self):
        # Test a simple THEMIS orbit plot in the XY plane, plot units in re
         tplotxy3('tha_pos_gse thd_pos_gse the_pos_gse',title='THEMIS inner orbit plot 2020-01-01, all planes, sun left, units re', legend_names=['THEMIS-A','THEMIS-D', 'THEMIS-E'], reverse_x=True, plot_units='re', display=global_display, save_png='orbits_thm_re_legends_3planes.png')

    def test_themis_art_orbit_re_legends_3planes(self):
        # Test a simple THEMIS orbit plot in the XY plane, plot units in re
         tplotxy3('th?_pos_gse',title='THEMIS_ARTEMIS orbit plot 2020-01-01, all planes, sun left, units re', legend_names=['THEMIS-A','THEMIS-D', 'THEMIS-E'], reverse_x=True, plot_units='re', display=global_display, save_png='orbits_thm_art_re_legends_3planes.png')

    def test_themis_orbit_km_legends_3planes_sunright_km(self):
        # Test a simple THEMIS orbit plot in the XY plane, plot units in re
        tplotxy3('tha_pos_gse thd_pos_gse the_pos_gse', title='THEMIS inner orbit plot 2020-01-01, all planes, sun right, units km',legend_names=['THEMIS-A','THEMIS-D', 'THEMIS-E'], plot_units='km', display=global_display, save_png='orbits_thm_re_legends_3planes_sunright_km.png')

    def test_themis_art_orbit_re_legends_extras_3planes(self):
        # Test a simple THEMIS orbit plot in the XY plane, plot units in re
        fig = tplotxy3('th?_pos_gse',title='THEMIS_ARTEMIS orbit plot 2020-01-01, all planes, extras, sun left, units re', legend_names=['THEMIS-A','THEMIS-B', 'THEMIS-C', 'THEMIS-D', 'THEMIS-E'], reverse_x=True, plot_units='re', display=False)
        # Add bow shock to figure
        tplotxy3_add_mpause(self.bs[0],self.bs[1],fig=fig,legend_name="Bow Shock",color='k',linestyle='dotted',linewidth=1,display=False)
        # Add magnetopause to figure
        tplotxy3_add_mpause(self.mp[0],self.mp[1],fig=fig,legend_name="Magnetopause",color='k',linestyle='dashed',linewidth=1,display=False)
        # Add neutral sheet to figure
        tplotxy3_add_neutral_sheet(self.gse_dat.y[:,0],self.gse_dat.y[:,2],fig=fig, legend_name="Neutral sheet",color='k',linestyle='dashdot',linewidth=1,display=global_display,save_png='orbits_thm_art_re_legends_extras_3planes.png')

    def test_themis_art_orbit_re_legends_extras_sunright_3planes(self):
        # Test a simple THEMIS orbit plot in the XY plane, plot units in re
        fig = tplotxy3('th?_pos_gse',title='THEMIS_ARTEMIS orbit plot 2020-01-01, all planes, extras, sun right, units re', legend_names=['THEMIS-A','THEMIS-B', 'THEMIS-C', 'THEMIS-D', 'THEMIS-E'], reverse_x=False, plot_units='re', display=False)
        # Add bow shock to figure
        tplotxy3_add_mpause(self.bs[0],self.bs[1],fig=fig,legend_name="Bow Shock",color='k',linestyle='dotted',linewidth=1,display=False)
        # Add magnetopause to figure
        tplotxy3_add_mpause(self.mp[0],self.mp[1],fig=fig,legend_name="Magnetopause",color='k',linestyle='dashed',linewidth=1,display=False)
        # Add neutral sheet to figure
        tplotxy3_add_neutral_sheet(self.gse_dat.y[:,0],self.gse_dat.y[:,2],fig=fig, legend_name="Neutral sheet",color='k',linestyle='dashdot',linewidth=1,display=global_display,save_png='orbits_thm_art_re_legends_extras_sunright_3planes.png')

    def test_themis_art_orbit_km_legends_extras_3planes(self):
        # Test a simple THEMIS orbit plot in the XY plane, plot units in re
        fig = tplotxy3('th?_pos_gse',title='THEMIS_ARTEMIS orbit plot 2020-01-01, all planes, extras, sun left, units km', legend_names=['THEMIS-A','THEMIS-B', 'THEMIS-C', 'THEMIS-D', 'THEMIS-E'],reverse_x=True, plot_units='km', display=False)
        # Add bow shock to figure
        tplotxy3_add_mpause(self.bs[0],self.bs[1],fig=fig,legend_name="Bow Shock",color='k',linestyle='dotted',linewidth=1,display=False)
        # Add magnetopause to figure
        tplotxy3_add_mpause(self.mp[0],self.mp[1],fig=fig,legend_name="Magnetopause",color='k',linestyle='dashed',linewidth=1,display=False)
        # Add neutral sheet to figure
        tplotxy3_add_neutral_sheet(self.gse_dat.y[:,0],self.gse_dat.y[:,2],fig=fig, legend_name="Neutral sheet",color='k',linestyle='dashdot',linewidth=1,display=global_display,save_png='orbits_thm_art_km_legends_extras_sunright_3planes.png')

    def test_plot_idl_traces(self):
        fig=tplotxy('tha_iono_t89_trace',display=global_display, plane='xy',plot_units='km',legend_names=['tha_t89_iono'],save_jpeg='flines_tha_t89_xy.jpeg')
        fig=tplotxy('tha_iono_t89_trace',display=global_display, plane='xz',plot_units='km',legend_names=['tha_t89_iono'],save_eps='flines_tha_t89_xy.eps')
        fig=tplotxy('tha_iono_t89_trace',display=global_display, plane='yz',plot_units='km',legend_names=['tha_t89_iono'],save_pdf='flines_tha_t89_yz.pdf')
        fig=tplotxy('ifoot_t89',display=global_display, plane='xy',plot_units='km',legend_names=['ifoot_t89'],save_svg='flines_ifoot_t89_xy.svg')
        fig=tplotxy('ifoot_t89',display=global_display, plane='xz',plot_units='km',legend_names=['ifoot_t89'],save_png='flines_ifoot_t89_xz.png')
        fig=tplotxy('ifoot_t89',display=global_display, plane='yz',plot_units='km',legend_names=['ifoot_t89'],save_png='flines_ifoot_t89_yz.png')

    def test_plot_idl_traces_3panels(self):
        fig=tplotxy3('tha_iono_t89_trace',display=global_display, plot_units='km',legend_names=['tha_t89_iono'],save_png='flines_tha_t89_3panel.png')
        fig=tplotxy3('ifoot_t89',display=global_display,plot_units='km',legend_names=['ifoot_t89'],save_png='flines_ifoot_t89_3panel.png')

if __name__ == "__main__":
    unittest.main()



import numpy as np

def bshock_2(xsh=None, *, short=False, xsh_max=14.3, return_west=False):
    """
    Compute location of bow shock, ported from the IDL SPEDAS bshock_2 utility

    Parameters
    ----------
    xsh : array-like or None
        If None or empty, generate xsh from xsh_min to xsh_max with npoints.
    short : bool
        If False (IDL default), return the full curve with east branch + mirrored west branch.
        If True, return only the east branch.
    xsh_max : float
        Maximum xsh used when generating xsh (IDL default 14.3).
    return_west : bool
        If True, also return ysh_west (on the original xsh grid, not mirrored/concatenated).

    Returns
    -------
    xsh_out : ndarray
    ysh_out : ndarray
    (optional) ysh_west : ndarray
    """
    # IDL defaults / constants
    xsh_min = -300.0
    npoints = 1000
    aberangle = 4.5  # present in IDL but unused in this routine

    # IDL: if n_elements(xsh) eq 0 then generate
    if xsh is None:
        xsh_arr = np.linspace(xsh_min, xsh_max, npoints, dtype=float)
    else:
        xsh_arr = np.asarray(xsh, dtype=float)
        if xsh_arr.size == 0:
            xsh_arr = np.linspace(xsh_min, xsh_max, npoints, dtype=float)

    # coefficients
    a1 = 0.2164
    b1 = -0.0986
    c1 = -4.26
    d1 = 44.916
    e1 = -623.77

    beta = a1 * xsh_arr + c1
    gamma = b1 * (xsh_arr ** 2) + d1 * xsh_arr + e1
    delta = beta**2 - 4.0 * gamma

    # Match IDL: sqrt(delta) (IDL would yield NaN/complex depending on settings;
    # NumPy will produce nan for negative delta with real dtype)
    sqrt_delta = np.sqrt(delta)

    ysh_east = (-beta - sqrt_delta) / 2.0
    ysh_west = (-beta + sqrt_delta) / 2.0

    if not short:
        # IDL:
        # ireverse = n_elements(xsh) - indgen(n_elements(xsh)) - 1
        # xsh = [xsh, xsh(ireverse)]
        # ysh = [ysh_east, ysh_west(ireverse)]
        x_rev = xsh_arr[::-1]
        y_out = np.concatenate([ysh_east, ysh_west[::-1]])
        x_out = np.concatenate([xsh_arr, x_rev])
    else:
        x_out = xsh_arr
        y_out = ysh_east

    if return_west:
        return x_out, y_out, ysh_west
    return x_out, y_out



from pyspedas.tplot_tools import tplot_wildcard_expand, get_data, get_coords, get_units
import logging
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Wedge
from .save_plot import save_plot

def tplotxy(tvars,
            plane='xy',
            center_origin=True,
            reverse_x = False,
            reverse_y = False,
            plot_units='re',
            title=None,
            colors=('k', 'b', 'g', 'r', 'c', 'm', 'y'),
            linestyles=('solid'),
            linewidths=(None),
            markers=(None),
            startmarkers=(None),
            endmarkers=(None),
            markevery=10,
            markersize=5,
            legend_names=None,
            legend_location='upper right',
            show_centerbody=True,
            centerbody_size_re=1.0,
            save_png='',
            save_eps='',
            save_jpeg='',
            save_pdf='',
            save_svg='',
            dpi=300,
            display=True,
            fig=None,
            axis=None,
            ):
    """
    Plot one or more 3d tplot variables, by projecting them onto one of the coordinate planes XY, XY, or YZ.

    Parameters
    ----------
    tvars: string or list of strings
        Tplot variables to be plotted (wildcards accepted).  The data array should be ntimesx3,
        or in the case of field line plots, ntimesxnpointsx3
    plane: string or list of strings
        Plane to project the 3d vectors onto. Valid options: 'xy', 'xz', 'yz'. Default: 'xy'
    center_origin: bool
        If True, center the plot on the origin.
    reverse_x: bool
        If True, reverse the x-axis of the plot (more positive values to the left).  This is handy
        for GSE-like orbit plots if you want the sun depicted toward the left
    reverse_y: bool
        If True, reverse the y-axis of the plot (more positive values to the bottom)
        If reverse_x is set, and you're projecting onto the xy plane, this keeps the coordinate system right-handed.
    colors: list or tuple of strings
        Colors to use if multiple variables are plotted. Default: ('k', 'b', 'g', 'r', 'c', 'm', 'y').
        If number of variables exceeds number of colors, they will cycle.
    linestyles: list or tuple of strings
        Line styles to use for each variable. Default: ('solid').
        If number of variables exceeds number of linestyles, they will cycle.
    linewidths: list or tuple of strings
        Line widths to use for each variable. If number of variables exceeds number of linewidths, they will cycle.
        Default: (None) (use matplotlib default)
    markers: list or tuple of str
        marker style for the data points (default: (None))
    startmarkers: list or tuple of str
        marker style for the start of each trace (default: (None))
    endmarkers: list or tuple of str
        marker style for the end of each trace (default: (None))
    markevery: int or sequence of int
        plot a marker at every n-th data point (default: 10)
    markersize: float
        size of the marker in points (default: 5)
    legend_names: list of str
        If set, labels to use in the plot legend.
    legend_location: str
        Placement of legend relative to the plot window.
    show_centerbody: bool
        If True, draw the central body at the origin
    centerbody_size_re: float
        The size in Re of the center body. Default: 1.0
    save_png : str, optional
        A full file name and path.
        If this option is set, the plot will be automatically saved to the file name provided in PNG format.
    save_eps : str, optional
        A full file name and path.
        If this option is set, the plot will be automatically saved to the file name provided in EPS format.
    save_jpeg : str, optional
        A full file name and path.
        If this option is set, the plot will be automatically saved to the file name provided in JPEG format.
    save_pdf : str, optional
        A full file name and path.
        If this option is set, the plot will be automatically saved to the file name provided in PDF format.
    save_svg : str, optional
        A full file name and path.
        If this option is set, the plot will be automatically saved to the file name provided in SVG format.
    dpi: float, optional
        The resolution of the plot in dots per inch
    display: bool, optional
        If True, then this function will display the plotted tplot variables. Use False to suppress display (for example, if
        saving to a file, or returning plot objects to be displayed later). Default: True
    fig: Matplotlib figure object
        Use an existing figure to plot in (mainly for recursive calls to render composite variables)
    axis: Matplotlib axes object
        Use an existing set of axes to plot on (mainly for recursive calls to render composite variables)


    Returns
    -------
    None
    """

    tvars = tplot_wildcard_expand(tvars)
    if len(tvars) < 1:
        logging.error("No matching variables found to plot")
        return

    plane = plane.lower()
    if plane not in ['xy', 'xz', 'yz']:
        logging.error(f'Invalid plane {plane}: must be one of "xy", "xz", "yz"')
        return
    # Get maximum range of each variable to set plot x/y range
    max_x = 0
    max_y = 0
    min_x = 0
    min_y = 0

    xsize=4.0
    ysize=4.0

    if not isinstance(colors, (list, np.ndarray,tuple)):
        colors = [colors]
    if not isinstance(markers, (list, np.ndarray,tuple)):
        markers = [markers]
    if not isinstance(startmarkers, (list, np.ndarray,tuple)):
        startmarkers = [startmarkers]
    if not isinstance(endmarkers, (list, np.ndarray,tuple)):
        endmarkers = [endmarkers]
    if not isinstance(linestyles, (list, np.ndarray,tuple)):
        linestyles = [linestyles]
    if not isinstance(linewidths, (list, np.ndarray,tuple)):
        linewidths = [linewidths]

    if legend_names is not None:
        if not isinstance(legend_names, (list, np.ndarray,tuple)):
            legend_names = [legend_names]
        if len(legend_names) != len(tvars):
            logging.warning(f"Length of legend_names ({len(legend_names)}) does not match length of tvars ({len(tvars)}), disabling legends")
            legend_names= None

    if fig is None and axis is None:
        fig, axis = plt.subplots(sharey=True, sharex=True, figsize=(xsize, ysize), layout='constrained')

    if title is not None and title != '':
        fig.suptitle(title)

    km_in_re = 6371.2
    if plot_units is None:
        plot_units = get_units(tvars[0])
    if isinstance(plot_units, (list, np.ndarray)):
        plot_units = plot_units[0]
    if plot_units is None:
        plot_units = 'None'
        unit_annotation=""
    else:
        unit_annotation=f' ({plot_units})'

    plot_coords = get_coords(tvars[0])
    if plot_coords is None:
        plot_coords = 'Unknown'
        coord_annotation=""
    else:
        coord_annotation=f"-{plot_coords}"

    full_annotation=coord_annotation+unit_annotation

    if plane=='xy':
        axis.set_xlabel('X' + full_annotation)
        axis.set_ylabel('Y' + full_annotation)

    elif plane=='xz':
        axis.set_xlabel('X' + full_annotation)
        axis.set_ylabel('Z' + full_annotation)

    elif plane=='yz':
        axis.set_xlabel('Y' + full_annotation)
        axis.set_ylabel('Z' + full_annotation)

    for index,tvar in enumerate(tvars):
        units=get_units(tvar)
        if isinstance(plot_units, (list, np.ndarray)):
            units=units[0]
        if units is None:
            units="None"
        unit_conv = 1.0
        if plot_units.lower() == 'km':
            if units.lower() == 'km':
                unit_conv = 1.0
            elif units.lower() == 're':
                unit_conv = km_in_re
            else:
                logging.warning(f"Input variable {tvar} has units {units}, unable to convert to plot_units {plot_units}")
        elif plot_units.lower() == 're':
            if units.lower() == 'km':
                unit_conv = 1.0/km_in_re
            elif units.lower() == 're':
                unit_conv = 1.0
            else:
                logging.warning(f"Input variable {tvar} has units {units.lower()}, unable to convert to plot_units {plot_units.lower()}")
        elif plot_units.lower() != units.lower():
                logging.warning(f"Input variable {tvar} has units {units.lower()}, unable to convert to plot_units {plot_units.lower}")
        d=get_data(tvar)
        ndim = len(d.y.shape)
        if ndim == 2: # orbits, foot points, hodograms, etc
            if plane=='xy':
                proj_x = d.y[:,0] * unit_conv
                proj_y = d.y[:,1] * unit_conv

            elif plane=='xz':
                proj_x = d.y[:,0] * unit_conv
                proj_y = d.y[:,2] * unit_conv

            elif plane=='yz':
                proj_x = d.y[:,1] * unit_conv
                proj_y = d.y[:,2] * unit_conv
        elif ndim == 3: # multiple field line traces
            if plane == 'xy':
                proj_x = d.y[:,:, 0] * unit_conv
                proj_y = d.y[:,:, 1] * unit_conv

            elif plane == 'xz':
                proj_x = d.y[:,:, 0] * unit_conv
                proj_y = d.y[:,:, 2] * unit_conv

            elif plane == 'yz':
                proj_x = d.y[:,:, 1] * unit_conv
                proj_y = d.y[:,:, 2] * unit_conv
        else:
            logging.error(f"Input variable {tvar} with {ndim} dimensions is not supported")
            return None

        local_xmax = np.nanmax(proj_x)
        local_ymax = np.nanmax(proj_y)
        local_xmin = np.nanmin(proj_x)
        local_ymin = np.nanmin(proj_y)

        max_x = np.nanmax([local_xmax, max_x])
        max_y = np.nanmax([local_ymax, max_y])
        min_x = np.nanmin([local_xmin, min_x])
        min_y = np.nanmin([local_ymin, min_y])

        n_colors = len(colors)
        thiscolor = colors[index % n_colors]
        n_styles = len(linestyles)
        thisstyle = linestyles[index % n_styles]
        n_widths = len(linewidths)
        thiswidth = linewidths[index % n_widths]
        n_markers = len(markers)
        thismarker = markers[index % n_markers]
        n_startmarkers = len(startmarkers)
        thisstartmarker = startmarkers[index % n_startmarkers]
        n_endmarkers = len(endmarkers)
        thisendmarker = endmarkers[index % n_endmarkers]

        this_line = axis.plot(proj_x, proj_y, color=thiscolor, linestyle=thisstyle, linewidth=thiswidth, marker=thismarker, markersize=markersize, markevery=markevery)
        if thisstartmarker is not None:
            axis.plot(proj_x[0], proj_y[0], color=thiscolor, linestyle=thisstyle, marker=thisstartmarker, markersize=markersize)
        if thisendmarker is not None:
            axis.plot(proj_x[-1], proj_y[-1], color=thiscolor, linestyle=thisstyle, marker=thisendmarker, markersize=markersize)

        if legend_names is not None:
            try:
                if isinstance(this_line, list):
                    this_line[0].set_label(legend_names[index])
                else:
                    this_line.set_label(legend_names[index])
            except IndexError:
                continue

    if center_origin:
        x_halfwidth = np.nanmax(np.abs([max_x, min_x]))
        y_halfwidth = np.nanmax(np.abs([max_y, min_y]))
        # Add 5% to each boundary to give a little margin
        all_halfwidth = 1.05 * np.nanmax([x_halfwidth, y_halfwidth])
        axis.set_xlim((-all_halfwidth, all_halfwidth))
        axis.set_ylim((-all_halfwidth, all_halfwidth))
    else:
        # Adjust the X and Y ranges by equal increments to give a little margin
        x_adjust = .05*(max_x - min_x)
        y_adjust = .05*(max_y - min_y)
        all_adjust = np.nanmax([x_adjust, y_adjust])
        xr = (min_x-all_adjust, max_x+all_adjust)
        yr = (min_y-all_adjust, max_y+all_adjust)
        axis.set_xlim(xr)
        axis.set_ylim(yr)

    if show_centerbody:
        if plot_units == 're':
            cb_radius = centerbody_size_re
        else:
            cb_radius = centerbody_size_re * km_in_re

        theta1, theta2 = -90, 90
        if plane=='xy' or plane=='xz':
            w1 = Wedge((0, 0), cb_radius, theta1, theta2, fc='white', edgecolor='black')
            w2 = Wedge((0, 0), cb_radius, theta2, theta1, fc='black', edgecolor='black')
            axis.add_artist(w1)
            axis.add_artist(w2)
        else:
            theta1,theta2 = 0,360
            w1 = Wedge((0, 0), cb_radius, theta1, theta2, fc='white', edgecolor='black')
            axis.add_artist(w1)

    axis.set_aspect('equal')
    if reverse_x:
        axis.invert_xaxis()
    if reverse_y:
        axis.invert_yaxis()

    if legend_names is not None:
        legend = axis.legend(loc=legend_location, markerfirst=True)

    fig.canvas.draw()

    save_plot(save_png=save_png, save_eps=save_eps, save_jpeg=save_jpeg, save_pdf=save_pdf, save_svg=save_svg, dpi=dpi)

    if display:
        plt.show()





from matplotlib import pyplot as plt

def save_plot(save_png=None, save_eps=None, save_svg=None, save_pdf=None, save_jpeg=None, dpi=300):

    if save_png is not None and save_png != '':
        if not save_png.endswith('.png'):
            save_png += '.png'
        plt.savefig(save_png, dpi=dpi)

    if save_eps is not None and save_eps != '':
        if not save_eps.endswith('.eps'):
            save_eps += '.eps'
        plt.savefig(save_eps, dpi=dpi)

    if save_svg is not None and save_svg != '':
        if not save_svg.endswith('.svg'):
            save_svg += '.svg'
        plt.savefig(save_svg, dpi=dpi)

    if save_pdf is not None and save_pdf != '':
        if not save_pdf.endswith('.pdf'):
            save_pdf += '.pdf'
        plt.savefig(save_pdf, dpi=dpi)

    if save_jpeg is not None and save_jpeg != '':
        if not save_jpeg.endswith('.jpeg'):
            save_jpeg += '.jpeg'
        plt.savefig(save_jpeg, dpi=dpi)



from pyspedas.tplot_tools import tplot_wildcard_expand, get_data, get_coords, get_units, set_coords, set_units, store_data
import logging
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.patches import Wedge
from .save_plot import save_plot

km_in_re = 6371.2

def tplotxy3_add_mpause(x,
                        y,
                        fig=None,
                        legend_name=None,
                        color='k',
                        linestyle='solid',
                        linewidth=1,
                        units='re',
                        display=False,
                        save_png='',
                        save_eps='',
                        save_jpeg='',
                        save_pdf='',
                        save_svg='',
                        dpi=300,
                        ):
    """
    Add a magnetopause boundary or similar structure to a tplotxy3 figure

    This utility adds a magnetopause, bow shock, or similar structure to the XZ and XY planes of a tplotxy3 figure.
    The assumptions are that the structure is rotationally symmetric about the X axis, and the boundary
    is given by arrays of X and Y coordinates.

    The plot is assumed to be in units of either km or re.

    The plot X/Y/Z ranges will not be affected when the data is plotted.

    Parameters
    ----------
    x: array of floats
        The X coordinates of the boundary, presumably in GSE or GSM coordinates.
    y: array of floats
        The Y coordinates of the boundary, presumably in GSE or GSM coordinates.
    fig:
        The matplotlib figure object containing the panels to be updated.
    legend_name: str
        The string to add to the legend.
    color: str
        The color of the line to be plotted
    linestyle: str
        The style of the line to be plotted (e.g. 'solid', 'dotted', etc)
    linewidth: int
        The width of the line to be plotted
    units: str
        The units of the boundary position (default: Re)
    display: bool
        If True, display the figure after updating. Default: False
    save_png : str, optional
        A full file name and path.
        If this option is set, the plot will be automatically saved to the file name provided in PNG format.
    save_eps : str, optional
        A full file name and path.
        If this option is set, the plot will be automatically saved to the file name provided in EPS format.
    save_jpeg : str, optional
        A full file name and path.
        If this option is set, the plot will be automatically saved to the file name provided in JPEG format.
    save_pdf : str, optional
        A full file name and path.
        If this option is set, the plot will be automatically saved to the file name provided in PDF format.
    save_svg : str, optional
        A full file name and path.
        If this option is set, the plot will be automatically saved to the file name provided in SVG format.
    dpi: float, optional
        The resolution of the plot in dots per inch

    Note
    ====

    If you intend to add other elements like neutral sheet or magnetopause boundaries, be sure to
    use display=False until the final element is added.  Otherwise, closing the plot window will
    destroy the matplotlib plot objects and cause blank plots.

    Returns
    -------
    None
    """

    if fig is None:
        logging.error("No figure provided")
        return

    # Get axes and properties stored in the fig object

    xy_plane = fig.xy_plane
    xz_plane = fig.xz_plane
    yz_plane = fig.yz_plane
    plot_units = fig.plot_units

    unit_conv = 1.0

    if units.lower() != plot_units.lower():
        if units.lower() == 're':
            unit_conv = km_in_re
        else:
            unit_conv = 1.0/km_in_re

    # Ensure that the axes don't autoscale when adding the new trace
    xy_plane.autoscale(False, axis='both')
    xz_plane.autoscale(False, axis='both')

    # Plot the trace on each plane

    xy_plane.plot(x*unit_conv, y*unit_conv, color=color, linestyle=linestyle, linewidth=linewidth)
    this_line = xz_plane.plot(x*unit_conv, y*unit_conv, color=color, linestyle=linestyle, linewidth=linewidth)

    if legend_name is not None and legend_name != '':
        fig.legend().remove()
        try:
            if isinstance(this_line, list):
                this_line[0].set_label(legend_name)
            else:
                this_line.set_label(legend_name)
        except IndexError:
            pass

        # Align the top of the legend box with the top of the XY subplot, and the right of the legend box with the right of
        # the YZ subplot

        bbox1 = xy_plane.get_position()
        bbox2 = yz_plane.get_position()

        legend_bbox_top = bbox1.y1
        legend_bbox_right = bbox2.x1

        handles, labels = xz_plane.get_legend_handles_labels()
        legend = fig.legend(handles, labels, loc="upper right", markerfirst=True, bbox_to_anchor=(legend_bbox_right, legend_bbox_top), framealpha=1.0)

    fig.canvas.draw()

    save_plot(save_png=save_png, save_eps=save_eps, save_jpeg=save_jpeg, save_pdf=save_pdf, save_svg=save_svg, dpi=dpi)

    if display:
        plt.show()

def tplotxy3_add_neutral_sheet( x,
                                y,
                                fig=None,
                                legend_name=None,
                                color='k',
                                linestyle='solid',
                                linewidth=1,
                                units='re',
                                display=False,
                                save_png='',
                                save_eps='',
                                save_jpeg='',
                                save_pdf='',
                                save_svg='',
                                dpi=300,
                                ):
    """
    Add a neutral or similar structure to a tplotxy3 figure

    This utility adds a neutral sheet or similar structure to the XZ plane of a tplotxy3 figure.
    The boundary is given by arrays of X and Z coordinates.

    The plot is assumed to be in units of either km or re.

    The plot X/Y/Z ranges will not be affected when the data is plotted.

    Parameters
    ----------
    x: array of floats
        The X coordinates of the boundary, presumably in GSE or GSM coordinates.
    y: array of floats
        The Y coordinates of the boundary, presumably in GSE or GSM coordinates.
    fig:
        The matplotlib figure object containing the panels to be updated.
    legend_name: str
        The string to add to the legend.
    color: str
        The color of the line to be plotted
    linestyle: str
        The style of the line to be plotted (e.g. 'solid', 'dotted', etc)
    linewidth: int
        The width of the line to be plotted
    units: str
        The units of the boundary position (default: Re)
    display: bool
        If True, display the figure after updating. Default: False
    save_png : str, optional
        A full file name and path.
        If this option is set, the plot will be automatically saved to the file name provided in PNG format.
    save_eps : str, optional
        A full file name and path.
        If this option is set, the plot will be automatically saved to the file name provided in EPS format.
    save_jpeg : str, optional
        A full file name and path.
        If this option is set, the plot will be automatically saved to the file name provided in JPEG format.
    save_pdf : str, optional
        A full file name and path.
        If this option is set, the plot will be automatically saved to the file name provided in PDF format.
    save_svg : str, optional
        A full file name and path.
        If this option is set, the plot will be automatically saved to the file name provided in SVG format.
    dpi: float, optional
        The resolution of the plot in dots per inch

    Note
    ====

    If you intend to add other elements like neutral sheet or magnetopause boundaries, be sure to
    use display=False until the final element is added.  Otherwise, closing the plot window will
    destroy the matplotlib plot objects and cause blank plots.

    Returns
    -------
    None
    """

    if fig is None:
        logging.error("No figure provided")
        return

    # Get axes and properties stored in the fig object

    xz_plane = fig.xz_plane
    xy_plane = fig.xy_plane # We will need this to properly position the legend box
    yz_plane = fig.yz_plane

    plot_units = fig.plot_units

    unit_conv = 1.0

    if units.lower() != plot_units.lower():
        if units.lower() == 're':
            unit_conv = km_in_re
        else:
            unit_conv = 1.0/km_in_re

    # Ensure that the axes don't autoscale when adding the new trace
    xz_plane.autoscale(False, axis='both')

    # Plot the trace on each plane

    this_line = xz_plane.plot(x*unit_conv, y*unit_conv, color=color, linestyle=linestyle, linewidth=linewidth)

    if legend_name is not None and legend_name != '':
        fig.legend().remove()
        try:
            if isinstance(this_line, list):
                this_line[0].set_label(legend_name)
            else:
                this_line.set_label(legend_name)
        except IndexError:
            pass

        # Align the top of the legend box with the top of the XY subplot, and the right of the legend box with the right of
        # the YZ subplot

        bbox1 = xy_plane.get_position()
        bbox2 = yz_plane.get_position()

        legend_bbox_top = bbox1.y1
        legend_bbox_right = bbox2.x1

        handles, labels = xz_plane.get_legend_handles_labels()
        legend = fig.legend(handles, labels, loc="upper right", markerfirst=True,bbox_to_anchor=(legend_bbox_right,legend_bbox_top),framealpha=1.0)

    fig.canvas.draw()

    save_plot(save_png=save_png, save_eps=save_eps, save_jpeg=save_jpeg, save_pdf=save_pdf, save_svg=save_svg, dpi=dpi)
    if display:
        plt.show()


def tplotxy3(tvars,
            center_origin=True,
            reverse_x = False,
            plot_units='re',
            title=None,
            colors=('k', 'b', 'g', 'r', 'c', 'm', 'y'),
            linestyles=('solid'),
            linewidths=(None),
            markers=(None),
            startmarkers=(None),
            endmarkers=(None),
            markevery=10,
            markersize=5,
            legend_names=None,
            show_centerbody=True,
            centerbody_size_re=1.0,
            save_png='',
            save_eps='',
            save_jpeg='',
            save_pdf='',
            save_svg='',
            dpi=300,
            display=True,
            fig=None,
            axis=None,
            ):
    """
    Plot one or more 3d tplot variables, by projecting them onto the three coordinate axes planes in a single figure.

    Parameters
    ----------
    tvars: string or list of strings
        Tplot variables to be plotted (wildcards accepted).  The data array should be ntimesx3,
        or in the case of field line plots, ntimesxnpointsx3
    center_origin: bool
        If True, center the plot on the origin.
    reverse_x: bool
        If True, reverse the x-axis of the plot (more positive values to the left).
        For plots in GSE-like coordinates, this puts the sun to the left instead of the right.
        It will also reverse the Y axis of the upper left (XY) plot.
    plot_units: str
        The units to use for the plot (default: 're').
    title: str
        The title to use for the plot.
    colors: list or tuple of strings
        Colors to use if multiple variables are plotted. Default: ('k', 'b', 'g', 'r', 'c', 'm', 'y').
        If number of variables exceeds number of colors, they will cycle.
    linestyles: list or tuple of strings
        Line styles to use for each variable. Default: ('solid').
        If number of variables exceeds number of linestyles, they will cycle.
    linewidths: list or tuple of strings
        Line widths to use for each variable. If number of variables exceeds number of linewidths, they will cycle.
        Default: (None) (use matplotlib default)
    markers: list or tuple of str
        marker style for the data points (default: (None))
    startmarkers: list or tuple of str
        marker style for the start of each trace (default: (None))
    endmarkers: list or tuple of str
        marker style for the end of each trace (default: (None))
    markevery: int or sequence of int
        plot a marker at every n-th data point (default: 10)
    markersize: float
        size of the marker in points (default: 5)
    legend_names: list of str
        If set, labels to use in the plot legend.
    show_centerbody: bool
        If True, draw the central body at the origin
    centerbody_size_re: float
        The size in Re of the center body. Default: 1.0
    save_png : str, optional
        A full file name and path.
        If this option is set, the plot will be automatically saved to the file name provided in PNG format.
    save_eps : str, optional
        A full file name and path.
        If this option is set, the plot will be automatically saved to the file name provided in EPS format.
    save_jpeg : str, optional
        A full file name and path.
        If this option is set, the plot will be automatically saved to the file name provided in JPEG format.
    save_pdf : str, optional
        A full file name and path.
        If this option is set, the plot will be automatically saved to the file name provided in PDF format.
    save_svg : str, optional
        A full file name and path.
        If this option is set, the plot will be automatically saved to the file name provided in SVG format.
    dpi: float, optional
        The resolution of the plot in dots per inch
    display: bool, optional
        If True, then this function will display the plotted tplot variables. Use False to suppress display (for example, if
        saving to a file, or returning plot objects to be displayed later). Default: True
    fig: Matplotlib figure object
        Use an existing figure to plot in (mainly for recursive calls to render composite variables)
    axis: Matplotlib axes object
        Use an existing set of axes to plot on (mainly for recursive calls to render composite variables)


    Note
    ====

    If you intend to add other elements like neutral sheet or magnetopause boundaries, be sure to
    use display=False until the final element is added.  Otherwise, closing the plot window will
    destroy the matplotlib plot objects and cause blank plots.

    Returns
    -------
    Matplotlib fig object
    """

    tvars = tplot_wildcard_expand(tvars)
    if len(tvars) < 1:
        logging.error("No matching variables found to plot")
        return None

    # Get maximum range of each variable to set plot x/y range
    max_x = 0
    max_y = 0
    min_x = 0
    min_y = 0
    max_z = 0
    min_z = 0

    xsize=4.0
    ysize=4.0

    if not isinstance(colors, (list, np.ndarray,tuple)):
        colors = [colors]
    if not isinstance(markers, (list, np.ndarray,tuple)):
        markers = [markers]
    if not isinstance(startmarkers, (list, np.ndarray,tuple)):
        startmarkers = [startmarkers]
    if not isinstance(endmarkers, (list, np.ndarray,tuple)):
        endmarkers = [endmarkers]
    if not isinstance(linestyles, (list, np.ndarray,tuple)):
        linestyles = [linestyles]
    if not isinstance(linewidths, (list, np.ndarray,tuple)):
        linewidths = [linewidths]

    if legend_names is not None:
        if not isinstance(legend_names, (list, np.ndarray,tuple)):
            legend_names = [legend_names]
        if len(legend_names) != len(tvars):
            logging.warning(f"Length of legend_names ({len(legend_names)}) does not match length of tvars ({len(tvars)}), disabling legends")
            legend_names= None

    if fig is None and axis is None:
        fig, axis = plt.subplots(nrows=2, ncols=2, sharey='row', sharex='col', figsize=(2.5*xsize, 2.5*ysize), constrained_layout=True)

    xy_plane = axis[0,0]
    xz_plane = axis[1,0]
    yz_plane = axis[1,1]
    blank_plane = axis[0,1]

    # Save the plottable axes in the fig object in case we need to modify the plot later
    fig.xy_plane = xy_plane
    fig.xz_plane = xz_plane
    fig.yz_plane = yz_plane

    if title is not None and title != '':
        fig.suptitle(title)

    if plot_units is None:
        plot_units = get_units(tvars[0])
    if isinstance(plot_units, (list, np.ndarray)):
        plot_units = plot_units[0]
    if plot_units is None:
        plot_units = 'None'
        unit_annotation=""
    else:
        unit_annotation=f' ({plot_units}) '

    fig.plot_units = plot_units

    plot_coords = get_coords(tvars[0])
    if plot_coords is None:
        plot_coords = 'Unknown'
        coord_annotation=""
    else:
        coord_annotation=f"-{plot_coords}"
    fig.plot_coords = plot_coords

    full_annotation=coord_annotation+unit_annotation

    xy_plane.set_xlabel('X' + full_annotation)
    xy_plane.set_ylabel('Y' + full_annotation)

    xz_plane.set_xlabel('X' + full_annotation)
    xz_plane.set_ylabel('Z' + full_annotation)

    yz_plane.set_xlabel('Y' + full_annotation)
    yz_plane.set_ylabel('Z' + full_annotation)

    for index,tvar in enumerate(tvars):
        units=get_units(tvar)
        if isinstance(plot_units, (list, np.ndarray)):
            units=units[0]
        if units is None:
            units="None"
        unit_conv = 1.0
        if plot_units.lower() == 'km':
            if units.lower() == 'km':
                unit_conv = 1.0
            elif units.lower() == 're':
                unit_conv = km_in_re
            else:
                logging.warning(f"Input variable {tvar} has units {units}, unable to convert to plot_units {plot_units}")
        elif plot_units.lower() == 're':
            if units.lower() == 'km':
                unit_conv = 1.0/km_in_re
            elif units.lower() == 're':
                unit_conv = 1.0
            else:
                logging.warning(f"Input variable {tvar} has units {units.lower()}, unable to convert to plot_units {plot_units.lower()}")
        elif plot_units.lower() != units.lower():
                logging.warning(f"Input variable {tvar} has units {units.lower()}, unable to convert to plot_units {plot_units.lower}")
        d=get_data(tvar)

        ndims = len(d.y.shape)
        if ndims == 2: # orbits, hodograms, etc
            proj_x = d.y[:,0] * unit_conv
            proj_y = d.y[:,1] * unit_conv
            proj_z = d.y[:,2] * unit_conv
        elif ndims == 3: # multiple field line traces
            proj_x = d.y[:,:,0] * unit_conv
            proj_y = d.y[:,:,1] * unit_conv
            proj_z = d.y[:,:,2] * unit_conv
        else:
            logging.error(f"Input variable {tvar} with {ndims} dimensions is not supported")
            return None


        local_xmax = np.nanmax(proj_x)
        local_ymax = np.nanmax(proj_y)
        local_zmax = np.nanmax(proj_z)
        local_xmin = np.nanmin(proj_x)
        local_ymin = np.nanmin(proj_y)
        local_zmin = np.nanmin(proj_z)

        max_x = np.nanmax([local_xmax, max_x])
        max_y = np.nanmax([local_ymax, max_y])
        max_z = np.nanmax([local_zmax, max_z])
        min_x = np.nanmin([local_xmin, min_x])
        min_y = np.nanmin([local_ymin, min_y])
        min_z = np.nanmin([local_zmin, min_z])

        n_colors = len(colors)
        thiscolor = colors[index % n_colors]
        n_styles = len(linestyles)
        thisstyle = linestyles[index % n_styles]
        n_widths = len(linewidths)
        thiswidth = linewidths[index % n_widths]
        n_markers = len(markers)
        thismarker = markers[index % n_markers]
        n_startmarkers = len(startmarkers)
        thisstartmarker = startmarkers[index % n_startmarkers]
        n_endmarkers = len(endmarkers)
        thisendmarker = endmarkers[index % n_endmarkers]

        # XY plane
        this_axis = xy_plane
        this_axis.plot(proj_x, proj_y, color=thiscolor, linestyle=thisstyle, linewidth=thiswidth, marker=thismarker, markersize=markersize, markevery=markevery)
        if thisstartmarker is not None:
            this_axis.plot(proj_x[0], proj_y[0], color=thiscolor, linestyle=thisstyle, marker=thisstartmarker, markersize=markersize)
        if thisendmarker is not None:
            this_axis.plot(proj_x[-1], proj_y[-1], color=thiscolor, linestyle=thisstyle, marker=thisendmarker, markersize=markersize)

        # XZ plane
        # If present, the neutral sheet will only be plotted on this axis. So it's the
        # one we'll use to track the legends.
        this_axis = xz_plane
        this_line = this_axis.plot(proj_x, proj_z, color=thiscolor, linestyle=thisstyle, linewidth=thiswidth, marker=thismarker, markersize=markersize, markevery=markevery)
        if thisstartmarker is not None:
            this_axis.plot(proj_x[0], proj_y[0], color=thiscolor, linestyle=thisstyle, marker=thisstartmarker, markersize=markersize)
        if thisendmarker is not None:
            this_axis.plot(proj_x[-1], proj_y[-1], color=thiscolor, linestyle=thisstyle, marker=thisendmarker, markersize=markersize)

        # YZ plane
        this_axis = yz_plane
        this_axis.plot(proj_y, proj_z, color=thiscolor, linestyle=thisstyle, linewidth=thiswidth, marker=thismarker, markersize=markersize, markevery=markevery)
        if thisstartmarker is not None:
            this_axis.plot(proj_x[0], proj_y[0], color=thiscolor, linestyle=thisstyle, marker=thisstartmarker, markersize=markersize)
        if thisendmarker is not None:
            this_axis.plot(proj_x[-1], proj_y[-1], color=thiscolor, linestyle=thisstyle, marker=thisendmarker, markersize=markersize)

        if legend_names is not None:
            try:
                if isinstance(this_line, list):
                    this_line[0].set_label(legend_names[index])
                else:
                    this_line.set_label(legend_names[index])
            except IndexError:
                continue

    # Adjust X, Y, Z ranges
    if center_origin:
        x_halfwidth = np.nanmax(np.abs([max_x, min_x]))
        y_halfwidth = np.nanmax(np.abs([max_y, min_y]))
        z_halfwidth = np.nanmax(np.abs([max_z, min_z]))
        # Add 5% to each boundary to give a little margin
        all_halfwidth = 1.05 * np.nanmax([x_halfwidth, y_halfwidth, z_halfwidth])
        axis[0,0].set_xlim(-all_halfwidth, all_halfwidth)
        axis[0,0].set_ylim(-all_halfwidth, all_halfwidth)
        axis[1,0].set_xlim(-all_halfwidth, all_halfwidth)
        axis[1,0].set_ylim(-all_halfwidth, all_halfwidth)
        axis[1,1].set_xlim(-all_halfwidth, all_halfwidth)
        axis[1,1].set_ylim(-all_halfwidth, all_halfwidth)
    else:
        # Adjust the X and Y ranges by equal increments to give a little margin
        x_adjust = .05*(max_x - min_x)
        y_adjust = .05*(max_y - min_y)
        z_adjust = .05*(max_z - min_z)
        all_adjust = np.nanmax([x_adjust, y_adjust, z_adjust])
        xr = (min_x-all_adjust, max_x+all_adjust)
        yr = (min_y-all_adjust, max_y+all_adjust)
        zr = (min_z-all_adjust, max_z+all_adjust)
        xy_plane.set_xlim(xr)
        xy_plane.set_ylim(yr)
        xz_plane.set_xlim(xr)
        xz_plane.set_ylim(zr)
        yz_plane.set_xlim(yr)
        yz_plane.set_ylim(zr)

    if show_centerbody:
        if plot_units == 're':
            cb_radius = centerbody_size_re
        else:
            cb_radius = centerbody_size_re * km_in_re

        theta1, theta2 = -90, 90

        # Artists can't be shared between axes, so we need two copies of each wedge (XY and XZ)
        w1 = Wedge((0, 0), cb_radius, theta1, theta2, fc='white', edgecolor='black')
        w2 = Wedge((0, 0), cb_radius, theta2, theta1, fc='black', edgecolor='black')
        w3 = Wedge((0, 0), cb_radius, theta1, theta2, fc='white', edgecolor='black')
        w4 = Wedge((0, 0), cb_radius, theta2, theta1, fc='black', edgecolor='black')

        # XY
        xy_plane.add_artist(w1)
        xy_plane.add_artist(w2)

        # XZ
        xz_plane.add_artist(w3)
        xz_plane.add_artist(w4)

        # YZ  (this is just a circle)
        theta1,theta2 = 0,360
        w5 = Wedge((0, 0), cb_radius, theta1, theta2, fc='white', edgecolor='black')
        yz_plane.add_artist(w5)

    xy_plane.set_aspect('equal')
    xz_plane.set_aspect('equal')
    yz_plane.set_aspect('equal')

    if reverse_x:
        xy_plane.invert_xaxis()  # This should also reverse X axis on XZ plot, since they're shared
        xy_plane.invert_yaxis()  # This keeps the coordinate system right-handed
        #xz_plane.invert_xaxis() # XY and XZ plots share an X axis, so we don't need another flip

    # Grab the legend handles created for the XZ plot, and graft them onto the figure
    # instead of showing them on the XZ panel.

    # We need to do a draw() to lay out all the elements, so we know where to put the legend
    blank_plane.axis('off')
    fig.canvas.draw()

    # Align the top of the legend box with the top of the XY subplot, and the right of the legend box with the right of
    # the YZ subplot

    bbox1 = xy_plane.get_position()
    bbox2 = yz_plane.get_position()

    legend_bbox_top = bbox1.y1
    legend_bbox_right = bbox2.x1

    handles, labels = xz_plane.get_legend_handles_labels()
    if legend_names is not None:
        legend = fig.legend(handles, labels, loc="upper right", markerfirst=True, bbox_to_anchor=(legend_bbox_right,legend_bbox_top))

    fig.canvas.draw()

    save_plot(save_png=save_png, save_eps=save_eps, save_jpeg=save_jpeg, save_pdf=save_pdf, save_svg=save_svg, dpi=dpi)
    if display:
        plt.show()

    return fig





import logging
import pyspedas
from pyspedas.tplot_tools import store_data, get_data, tnames, get_units, set_units


def tkm2re(name, km=False, newname=None, suffix=''):
    """
    Converts a tplot variable to Re or Km

    Parameters
    -----------
        name: str or list of str
            Names of tplot variables to convert (wildcards accepted)

        km: bool, optional
            If True, convert Re to Km. If False, convert Km to Re.
            Default: False

        newname: str or list of str, optional
            Output variable names; if not set, the names of the
            input variables are used + '_re' or '_km'
            Default: None

        suffix: str, optional
            Specify to append a suffix to each variable 
            (only valid if newname is not specified)
            Default: ''

    Returns
    --------
        List of the tplot variables created

    Examples
    --------
        >>> import pyspedas
        >>> pyspedas.store_data('a', data={'x':[0,4,8,12,16], 'y':[1,2,3,4,5]})
        >>> pyspedas.tkm2re('a', km = True)


    """
    km_in_re = 6371.2

    names = tnames(name)

    if names == []:
        logging.error('tkm2re: No valid tplot variables found: ' + name)
        return

    if newname is None:
        newname = [n + suffix for n in names]

        if km == False:
            newname = [n + '_re' for n in newname]
        else:
            newname = [n + '_km' for n in newname]
    else:
        if not isinstance(newname, list):
            newname = [newname]

        if len(newname) != len(names):
            logging.error('Number of output variable names (newname) should match the number of input variables.')
            return

    out = []

    for in_tvar, out_tvar in zip(names, newname):
        data = get_data(in_tvar)
        metadata = get_data(in_tvar, metadata=True)

        if data is None:
            logging.error('Problem reading variable: ' + in_tvar)
            continue
        data_in_units = get_units(in_tvar)
        if data_in_units is None:
            data_in_units = 'None'
        if km == False:
            new_units = 'Re'
            if 're' in data_in_units.lower():
                logging.warning(f'Tplot variable {in_tvar} is already in units of Re!')
                data_out = data.y
            else:
                data_out = data.y/km_in_re
        else:
            new_units = 'km'
            if 'km' in data_in_units.lower():
                logging.warning(f'Tplot variable {in_tvar} is already in units of km!')
                data_out = data.y
            else:
                data_out = data.y*km_in_re

        saved = store_data(out_tvar, data={'x': data.times, 'y': data_out}, attr_dict=metadata)

        if not saved:
            logging.error('Problem creating tplot variable.')
            continue
        else:
            set_units(out_tvar, new_units)

        # update the subtitle, if needed
        yaxis_opt = pyspedas.tplot_tools.data_quants[out_tvar].attrs['plot_options'].get('yaxis_opt')

        if yaxis_opt is not None:
            subtitle = yaxis_opt.get('axis_subtitle')
            if subtitle is not None:
                if km == False:
                    new_subtitle = pyspedas.tplot_tools.data_quants[out_tvar].attrs['plot_options']['yaxis_opt']['axis_subtitle'].lower().replace('km', 'Re')
                else:
                    new_subtitle = pyspedas.tplot_tools.data_quants[out_tvar].attrs['plot_options']['yaxis_opt']['axis_subtitle'].lower().replace('re', 'km')
                pyspedas.tplot_tools.data_quants[out_tvar].attrs['plot_options']['yaxis_opt']['axis_subtitle'] = new_subtitle

        out.append(out_tvar)

    return out



import logging
import re
import numpy as np
from pyspedas.tplot_tools import get_data

logging.captureWarnings(True)
logging.basicConfig(format='%(asctime)s: %(message)s', datefmt='%d-%b-%y %H:%M:%S', level=logging.INFO)

def cluster_get_codif_dist(tname, probe=None):
    """
    Returns 3D particle data structures containing Cluster CODIF
    data for use with pySPEDAS particle routines. 
    
    Input
    ----------
        tname: str
            tplot variable name containing the CODIF distribution data
            Example: 'ions_3d__C1_CP_CIS_CODIF_HS_H1_PSD'

    Parameters
    ----------
        probe: str
            Spacecraft probe #

    Returns
    ----------
        3D particle data structure(s) containing Cluster CODIF distribution functions
    """
    data = get_data(tname)

    if data is None:
        logging.error('Problem extracting the CODIF distribution data.')
        return
    
    if any(data[i].ndim == 2 for i in (2, 3, 4)):
        raise ValueError(
            "data[2], data[3], and data[4] must all be 1-D arrays"
        )
    
    # check which species by looking at second last part of input string
    # this could be a function
    parts = tname.split("_")
    if re.fullmatch(r"[A-Z][a-z]?\d", parts[-2]):
        species = parts[-2].lower()
    else:
        raise ValueError(f"No valid species found in: {tname}")

    # possible masses in amu
    if species == 'h1':
        mass = 1.04535e-2 
        charge = 1.0
    elif species == 'he1':
        mass = 4.18138e-2
        charge = 1.0
    elif species == 'o1':
        mass = 0.167255
        charge = 1.0
    else:
        logging.error('Invalid species: ' + species)
        #return

    out = {'project_name': 'Cluster',
            'spacecraft': probe, 
            'species': species,
            'data_name': 'CODIF ' + species,
            'charge': charge,
            'units_name': 'df_km', 
            'mass': mass}

    # --- Reformat data
    # Shuffle the output to be [time, energy, phi, theta] 
    # (data[1] is currently [time, theta, phi, energy])
    out_data = data[1].transpose([0, 3, 2, 1]) #* 1.0E-18 * 1.0E-12 # convert from s^3 km^-6 to s^3 cm^-6
    out_bins = np.zeros(out_data.shape) + 1
    out_dphi = np.zeros(out_data.shape) + 22.5   # 16 pixels in 360 degree window
    out_dtheta = np.zeros(out_data.shape) + 22.5 # 8 pixels in 180 degree window
    out_denergy = np.zeros(out_data.shape) 

    # theta
    # elevations are constant across time
    theta_reform = np.asarray(data[2])                 # shape: (theta,)
    theta_reform = theta_reform[None, None, :]         # shape: (1, 1, theta)
    theta_rebinned = np.broadcast_to(
        theta_reform,
        (len(data[4]), len(data[3]), theta_reform.shape[2])
    )                                                  # shape: (energy, phi, theta) 
    out_theta = np.broadcast_to(
        theta_rebinned,
        (len(data[0]),) + theta_rebinned.shape
    )                                                  # shape: (time, energy, phi, theta) 

    # energy
    energy_len = len(data[4])
    energy_reform = np.asarray(data[4])[:, None, None] # shape: (energy, 1, 1)
    energy_table = np.broadcast_to(
        energy_reform,
        (len(data[4]), len(data[3]), len(data[2]))
    )                                                 # shape: (energy, phi, theta)
    out_energy = np.broadcast_to(
        energy_table,
        (len(data[0]),) + energy_table.shape
    )                                                 # shape: (time, energy, phi, theta)

    # phi
    phi_reform = np.asarray(data[3])[None, :, None]   # shape: (1, phi, 1)
    phi_rebinned = np.broadcast_to(
        phi_reform,
        (len(data[4]), len(data[3]), len(data[2])) 
    )                                                 # shape: (energy, phi, theta)
    out_phi = np.broadcast_to(
        phi_rebinned,
        (len(data[0]),) + phi_rebinned.shape
    )                                                 # shape: (time, energy, phi, theta)

    out_list = []
    for time_idx, time in enumerate(data[0]):
        out_table = {**out}
        out_table['data'] = out_data[time_idx, :]
        out_table['bins'] = out_bins[time_idx, :]
        out_table['theta'] = out_theta[time_idx, :]
        out_table['phi'] = out_phi[time_idx, :]
        out_table['energy'] = out_energy[time_idx, :]
        out_table['dtheta'] = out_dtheta[time_idx, :]
        out_table['dphi'] = out_dphi[time_idx, :]
        out_table['denergy'] = out_denergy[time_idx, :]
        out_table['n_energy'] = energy_len
        out_table['n_theta'] = len(data[3])
        out_table['n_phi'] = len(data[2])
        out_table['start_time'] = time # note: assumes the data weren't centered
        out_table['end_time'] = time + 4 # one spin = 4 seconds
        out_list.append(out_table)

    return out_list